﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEditor;

#if UNITY_EDITOR
/// <summary>
/// Toggles a variable's visibility based on a boolean or interger value (less than).
/// </summary>
public class VariableToggle : PropertyAttribute {
	
	public readonly string variableName;
	public readonly bool hideByBool = false;
	public readonly bool hideByInt = false;
	public readonly bool hideBool = false;
	public readonly int hideInt = 0;

	/// <summary>
	/// Turns off this variable's visibility if the the boolean variable matches the set boolean state.
	/// </summary>
	/// <param name="varName">Variable name.</param>
	/// <param name="state">If set to <c>true</c> state.</param>
	public VariableToggle (string varName, bool state) {
		this.variableName = varName;
		this.hideBool = state;
		this.hideByBool = true;
	}

	/// <summary>
	/// Turns off this variable's visibility if the the integer variable is less than the set integer value.
	/// </summary>
	/// <param name="varName">Variable name.</param>
	/// <param name="val">Value.</param>
	public VariableToggle (string varName, int val) {
		this.variableName = varName;
		this.hideInt = val;
		this.hideByInt = true;
	}
}

[CustomPropertyDrawer(typeof(VariableToggle))]
public class VariableToggleDrawer : PropertyDrawer {
	
	public override void OnGUI(Rect position, SerializedProperty property, GUIContent label) {
		if (!HideVariable (property)) {
			EditorGUI.PropertyField (position, property, true);
		}
	}

	public override float GetPropertyHeight (SerializedProperty property, GUIContent label) {
		if (HideVariable (property)) {
			return 0;
		} else {
			return base.GetPropertyHeight (property, label);
		}
	}

	private bool HideVariable (SerializedProperty property) {
		VariableToggle toggle = attribute as VariableToggle;
		SerializedProperty varToCheck = property.serializedObject.FindProperty (toggle.variableName);

		if (toggle.hideByBool)
			return varToCheck.boolValue == toggle.hideBool;
		else if (toggle.hideByInt)
			return varToCheck.intValue < toggle.hideInt;
		else
			return true;


	}
}

#else
public class VariableToggle : System.Attribute{
    public VariableToggle (string varName, int val) {}
    public VariableToggle (string varName, bool state) {}
}

public class VariableToggleDrawer : System.Attribute {}

#endif
